package de.renew.formalism.java;

import de.renew.expression.LocalVariable;

import de.renew.util.ClassSource;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;


public class ParsedDeclarationNode {
    List imports;
    Map wellKnownClasses;
    Map variables;
    private boolean anyVariablesDeclared;

    public ParsedDeclarationNode() {
        imports = new ArrayList();
        wellKnownClasses = new HashMap();
        variables = new HashMap();
        anyVariablesDeclared = false;
    }

    public void addImport(ParsedImportDecl decl, Token token)
            throws ParseException {
        if (decl.star) {
            // Remember this qualification for lookup.
            // Which imports should have priority: early or late imports?
            imports.add(decl.name);
        } else {
            // Calculate the short name of the class.
            int pos = decl.name.lastIndexOf(".");
            String name = decl.name.substring(pos + 1);

            // Warn about a name clash, if necessary.
            Object interpretation = interpreteWellKnownName(name);
            if (interpretation != null) {
                throw JavaHelper.makeParseException("Class " + name
                                                    + " imported twice.", token);
            }


            // Map the short name to the fully qualified name.
            wellKnownClasses.put(name, decl.name);
        }
    }

    public void addVariable(Class clazz, String name, Token token)
            throws ParseException {
        // Warn about a name clash, if necessary.
        Object interpretation = interpreteWellKnownName(name);
        if (interpretation != null) {
            if (interpretation instanceof LocalVariable) {
                throw JavaHelper.makeParseException("Variable " + name
                                                    + " declared twice.", token);
            } else {
                throw JavaHelper.makeParseException("Variable " + name
                                                    + " is named identically to an imported class.",
                                                    token);
            }
        }


        // Remember the variable.
        variables.put(name, clazz);

        if (!"this".equals(name)) {
            anyVariablesDeclared = true;
        }
    }

    private static boolean classExists(String name) {
        try {
            ClassSource.classForName(name);
            return true;
        } catch (Exception e) {
            return false;
        } catch (NoClassDefFoundError e) {
            return false;
        }
    }

    private static Class classForName(String name) {
        try {
            return ClassSource.classForName(name);
        } catch (ClassNotFoundException e) {
            throw new RuntimeException("Could not load class.");
        }
    }

    public Object interpreteWellKnownName(String name) {
        // Lookup the declaration. A variable?
        if (variables.containsKey(name)) {
            return new LocalVariable(name);
        }

        // An imported class?
        if (wellKnownClasses.containsKey(name)) {
            String iName = (String) wellKnownClasses.get(name);
            if (classExists(iName)) {
                return classForName(iName);
            }
        }
        return null;
    }

    // Return Class if class is denoted.
    // Return LocalVariable if variable is denoted.
    // Return null if no connection can be made.
    public Object interpreteName(String name) {
        // Is it a qualified name?
        if (name.indexOf(".") >= 0) {
            // This could be a qualified class.
            if (classExists(name)) {
                return classForName(name);
            }

            // Don't know. Maybe it is a package specification.
            return null;
        }

        // Is the object well-known?
        Object interpretation = interpreteWellKnownName(name);
        if (interpretation != null) {
            // Yes, the object was declared explicitly.
            return interpretation;
        }

        // A class in an imported package?
        Iterator packages = imports.iterator();
        while (packages.hasNext()) {
            String pckg = (String) packages.next();
            if (classExists(pckg + "." + name)) {
                return classForName(pckg + "." + name);
            }
        }

        // A class without package?
        if (classExists(name)) {
            return classForName(name);
        }

        // No. A language class?
        if (classExists("java.lang." + name)) {
            return classForName("java.lang." + name);
        }

        // Do we have any declarations?
        if (!anyVariablesDeclared) {
            // It is a variable, even though it is not declared.
            return new LocalVariable(name);
        } else {
            // Don't know. Maybe it is a package specification.
            return null;
        }
    }

    public Class findType(LocalVariable var) {
        if (variables.containsKey(var.name)) {
            return (Class) variables.get(var.name);
        } else {
            return de.renew.util.Types.UNTYPED;
        }
    }
}