package de.renew.gui;

import de.renew.application.SimulationEnvironment;
import de.renew.application.SimulatorPlugin;

import de.renew.plugin.PluginManager;
import de.renew.plugin.PluginProperties;

import java.awt.Component;

import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Vector;

import javax.swing.JOptionPane;


/**
 * Manages a configuration dialog for the Renew simulation. The
 * configuration is based on and affects the properties of the
 * Renew Simulator plugin.
 * <p>
 * Several theme-specific option tabs can be added. Each tab has
 * to be managed by a {@link ConfigureSimulationTabController}
 * implementation.
 * </p>
 *
 * @author Michael Duvigneau
 * @since Renew 2.0
 */
class ConfigureSimulationController {

    /**
     * Refers to the Swing dialog managed by this controller.
     **/
    private ConfigureSimulationDialog dialog;

    /**
     * Contains all {@link ConfigureSimulationTabController}
     * objects belonging to this dialog. The order in this
     * list matches the order of tabs in the dialog.
     **/
    private List tabControllers = new Vector();

    /**
     * Creates the controller and its associated dialog.
     * The frame of the given application is used as parent
     * for the dialog frame. Some well-known option tabs
     * are added to the dialog as well.
     *
     * @param app the application where the dialog should
     *            belong to.
     **/
    public ConfigureSimulationController(CPNApplication app) {
        tabControllers.add(new ConfigureEngineController());
        tabControllers.add(new ConfigureRemoteAccessController());
        tabControllers.add(new ConfigureNetpathController());
        Component[] initialTabs = new Component[tabControllers.size()];
        for (int i = 0; i < initialTabs.length; i++) {
            initialTabs[i] = ((ConfigureSimulationTabController) tabControllers
                              .get(i)).getTab();
        }
        this.dialog = new ConfigureSimulationDialog(app.getFrame(), this,
                                                    initialTabs);
    }

    /**
     * Adds an option tab to the dialog.
     * The panel controller is automatically instructed to update itself.
     *
     * @param controller the controller of the new option tab.
     **/
    public void addTabController(ConfigureSimulationTabController controller) {
        dialog.addOptionsTab(controller.getTab());
        tabControllers.add(controller);
        controller.updateTab(getSimulationProperties());
    }

    /**
     * Removes an option tab from the dialog.
     *
     * @param controller the controller of the option tab to remove.
     **/
    public void removeTabController(ConfigureSimulationTabController controller) {
        if (tabControllers.remove(controller)) {
            dialog.removeOptionsTab(controller.getTab());
        }
    }

    /**
     * Displays the dialog. Also updates the dialog, if it has not been
     * visible beforehand.
     **/
    public void showDialog() {
        if (!dialog.isVisible()) {
            updateDialog();
            dialog.setVisible(true);
        }
    }

    /**
     * Hides the dialog from view.
     **/
    public void closeDialog() {
        dialog.setVisible(false);
    }

    /**
     * Instructs all option tab controllers to write their
     * current settings to the simulator plugin's properties.
     * This method is intended to be called upon user request
     * only.
     **/
    public void commitDialog() {
        Properties props = getSimulatorPluginProperties();
        Iterator tabs = tabControllers.iterator();
        while (tabs.hasNext()) {
            ((ConfigureSimulationTabController) tabs.next()).commitTab(props);
        }
    }

    /**
     * Instructs all option tab controllers to bring their
     * current settings in line with the simulator plugin's
     * properties.
     * This method is intended to be called upon user request
     * or upon dialog initialisation.
     **/
    public void updateDialog() {
        Properties props = getSimulatorPluginProperties();
        Iterator tabs = tabControllers.iterator();
        while (tabs.hasNext()) {
            ((ConfigureSimulationTabController) tabs.next()).updateTab(props);
        }
    }

    /**
     * Instructs all option tab controllers to bring their
     * current settings in line with the current simulation's
     * properties.
     * This method is intended to be called upon user request
     * only.
     **/
    public void updateDialogFromSimulation() {
        Properties props = getSimulationProperties();
        if (props != null) {
            Iterator tabs = tabControllers.iterator();
            while (tabs.hasNext()) {
                ((ConfigureSimulationTabController) tabs.next()).updateTab(props);
            }
        } else {
            JOptionPane.showMessageDialog(dialog,
                                          "Could not update properties from current simulation:\n"
                                          + "No simulation running.",
                                          "Configure Simulation",
                                          JOptionPane.ERROR_MESSAGE);
        }
    }

    /**
     * Fetches the current properties of the Renew Simulator plugin.
     *
     * @return the properties object of the simulator plugin.
     **/
    private static Properties getSimulatorPluginProperties() {
        if (PluginProperties.debug()) {
            System.out.println("ConfigureSimulationController: "
                               + "Reading the simulator plugin's properties.");
        }
        return getSimulatorPlugin().getProperties();
    }

    /**
     * Fetches the properties of the current simulation
     * environment. If there is no running simulation,
     * <code>null</code> is returned instead.
     *
     * @return the properties object of the current simulation or
     *         the <code>null</code>, depending on whether there
     *         is a running simulation.
     **/
    private static Properties getSimulationProperties() {
        SimulatorPlugin simulatorPlugin = getSimulatorPlugin();
        Properties props;
        SimulationEnvironment env = simulatorPlugin.getCurrentEnvironment();
        if (env != null) {
            if (PluginProperties.debug()) {
                System.out.println("ConfigureSimulationController: "
                                   + "Reading the current simulation's properties.");
            }
            props = env.getProperties();
        } else {
            if (PluginProperties.debug()) {
                System.out.println("ConfigureSimulationController: "
                                   + "No current simulation available to read properties.");
            }
            props = null;
        }
        return props;
    }

    /**
     * Determines the Renew Simulator plugin instance.
     *
     * @return the first plugin providing <code>de.renew.simulator</code>.
     */
    private static SimulatorPlugin getSimulatorPlugin() {
        return (SimulatorPlugin) PluginManager.getInstance()
                                              .getPluginsProviding("de.renew.simulator")
                                              .iterator().next();
    }
}