package de.renew.plugin;

import de.renew.plugin.DependencyCheckList.DependencyElement;
import de.renew.plugin.command.CLCommand;
import de.renew.plugin.command.ExitCommand;
import de.renew.plugin.command.GCCommand;
import de.renew.plugin.command.GetPropertyCommand;
import de.renew.plugin.command.InfoCommand;
import de.renew.plugin.command.ListCommand;
import de.renew.plugin.command.LoadCommand;
import de.renew.plugin.command.NoOpCommand;
import de.renew.plugin.command.ScriptCommand;
import de.renew.plugin.command.SetPropertyCommand;
import de.renew.plugin.command.UnloadCommand;
import de.renew.plugin.load.PluginLoaderComposition;
import de.renew.plugin.load.SimplePluginLoader;
import de.renew.plugin.locate.PluginJarLocationFinder;
import de.renew.plugin.locate.PluginLocationFinders;
import de.renew.plugin.locate.PluginSubDirFinder;

import java.io.File;

import java.net.MalformedURLException;
import java.net.URL;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeMap;
import java.util.Vector;


/**
 * This class is the central management facility for providing
 * Plugins with access to Renew.
 * When the Singleton instance of this class is created, the statical plugins will be loaded.
 * Furthermore, this class will be the interface for
 * dynamical plugins that can log into it at run-time.
 */
public class PluginManager {

    /**
    * The timeout for cleaning up a plugin, in microseconds.
    * The value is fixed to {@value} microseconds.
    **/
    public static final int CLEANUP_TIMEOUT = 20000;

    /** Singleton member instance. **/
    private static PluginManager _instance;

    /** The list managing the dependencies among the plugins. **/
    DependencyCheckList _dependencyList = new DependencyCheckList();

    /**
     * The map of recognized commands and their associated executors.
     * Maps from <code>String</code> to {@link CLCommand} objects.
     **/
    private Map _commands = Collections.synchronizedMap(new TreeMap());

    /**
     * Contains the listeners that register for PluginManager events.
     */
    private Set _managerListener;

    /**
     * The set of plugins preventing the system from automatical
     * termination.
     **/
    private Set _blockers = Collections.synchronizedSet(new HashSet());

    /**
     * This flag indicates that the system termination thread is
     * running. It is set by the {@link #stopSynchronized()}
     * method only.
     **/
    private boolean isTerminating = false;
    private PluginLocationFinders _locationFinder = PluginLocationFinders
                                                    .getInstance();
    private PluginLoaderComposition _loader = new PluginLoaderComposition();
    private PluginClassLoader _pluginClassLoader;
    private ClassLoader _bottomClassLoader;
    private ClassLoader _systemClassLoader;

    protected PluginManager() {
        _managerListener = new HashSet();

        // create the PluginClassLoader
        _pluginClassLoader = new PluginClassLoader(getClass().getClassLoader());

        // get the SystemClassLoader
        _systemClassLoader = ClassLoader.getSystemClassLoader();

        // create the default BottomClassLoader
        _bottomClassLoader = new BottomClassLoader(_pluginClassLoader);

        initLocationFinders();

        // initialize the loaders
        _loader = new PluginLoaderComposition();
        _loader.addLoader(new SimplePluginLoader());

        _commands.put("", new NoOpCommand());
        _commands.put("help", new HelpCommand());
        _commands.put("exit", new ExitCommand());
        _commands.put("load", new LoadCommand());
        _commands.put("list", new ListCommand());
        _commands.put("unload", new UnloadCommand());
        _commands.put("info", new InfoCommand());
        _commands.put("script", new ScriptCommand());
        _commands.put("set", new SetPropertyCommand());
        _commands.put("get", new GetPropertyCommand());
        _commands.put("gc", new GCCommand());
        _commands.put("packageCount",
                      _pluginClassLoader.new PackageCountCommand());
    }

    public static synchronized PluginManager getInstance() {
        if (_instance == null) {
            _instance = new PluginManager();
            _instance.initPlugins();
        }
        return _instance;
    }

    /**
     * Adds a Listener to the PluginManager
     * @param l the listener
     */
    public void addPluginManagerListener(IPluginManagerListener l) {
        _managerListener.add(l);
    }

    /**
     * Removes a Listener from the PluginManager
     * @param l the listener
     */
    public void removePluginManagerListener(IPluginManagerListener l) {
        _managerListener.remove(l);
    }

    private void serviceAdded(Collection c) {
        for (Iterator iter = _managerListener.iterator(); iter.hasNext();) {
            IPluginManagerListener l = (IPluginManagerListener) iter.next();
            for (Iterator iterator = c.iterator(); iterator.hasNext();) {
                l.serviceAdded((String) iterator.next());
            }
        }
    }

    private void serviceRemoved(Collection c) {
        for (Iterator iter = _managerListener.iterator(); iter.hasNext();) {
            IPluginManagerListener l = (IPluginManagerListener) iter.next();
            for (Iterator iterator = c.iterator(); iterator.hasNext();) {
                l.serviceRemoved((String) iterator.next());
            }
        }
    }

    public void addPlugin(IPlugin plugin)
            throws DependencyNotFulfilledException {
        if (!checkDependenciesFulfilled(plugin)) {
            throw new DependencyNotFulfilledException("Cannot add " + plugin);
        }

        try {
            if (PluginProperties.debug()) {
                System.out.println("************ INITIALIZING "
                                   + plugin.getName() + " ********");
            }
            plugin.init();
            _dependencyList.addElement(new DependencyElement(plugin));
            serviceAdded(plugin.getProperties().getProvisions());
        } catch (RuntimeException e) {
            System.err.println("PluginManager: adding of Plugin " + plugin
                               + " failed: " + e);
            if (PluginProperties.debug()) {
                e.printStackTrace();
            }
        } catch (NoClassDefFoundError e) {
            System.err.println("PluginManager: adding of Plugin " + plugin
                               + " failed: " + e);
            if (PluginProperties.debug()) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Checks whether the dependencies specified by the given
     * <code>PluginProperties</code> object are fulfilled by the current
     * set of loaded plugins.
     *
     * @param props  the properties that specify the dependencies.
     *
     * @return <code>true</code>, if a plugin with the given <code>props</code>
     *         could be added to the set of plugins. Returns <code>false</code>,
     *         if some dependency of the plugin would not be fulfilled.
     **/
    public synchronized boolean checkDependenciesFulfilled(PluginProperties props) {
        return _dependencyList.dependencyFulfilled(new DependencyElement(props));
    }

    /**
     * Checks whether the dependencies specified by the given
     * <code>IPlugin</code> object are fulfilled by the current
     * set of loaded plugins.
     *
     * @param plugin  the plugin that specifies the dependencies.
     *
     * @return <code>true</code>, if the given <code>plugin</code> could
     *         be added to the set of plugins. Returns <code>false</code>,
     *         if some dependency of the plugin would not be fulfilled.
     **/
    public synchronized boolean checkDependenciesFulfilled(IPlugin plugin) {
        return _dependencyList.dependencyFulfilled(new DependencyElement(plugin));
    }

    /**
     * Returns the plugin with the given name, null if
     * no such plugin is present.
     */
    public IPlugin getPluginByName(String pluginName) {
        IPlugin found = null;
        if (PluginProperties.debug()) {
            System.out.println("PluginManager looking for " + pluginName);
        }
        List plugins = getPlugins();
        for (int i = 0; i < plugins.size(); i++) {
            IPlugin current = (IPlugin) plugins.get(i);
            if (current.getName().equals(pluginName)) {
                found = current;
                break;
            }
        }
        return found;
    }

    /**
     * Returns a Collection containing the Plugins that provide
     * the given service.
     */
    public Collection getPluginsProviding(String service) {
        Collection result = new Vector();
        Iterator it = getPlugins().iterator();
        while (it.hasNext()) {
            IPlugin p = (IPlugin) it.next();
            if (p.getProperties().getProvisions().contains(service)) {
                result.add(p);
            }
        }
        return result;
    }

    /**
     * Iterate through all plugins and initialize them.
     */
    public void initPlugins() {
        Iterator plugins = getPlugins().iterator();

        while (plugins.hasNext()) {
            IPlugin p = (IPlugin) plugins.next();
            if (PluginProperties.debug()) {
                System.out.println("initializing " + p.getName()
                                   + ", loaded from "
                                   + p.getClass().getClassLoader());
            }
            if (PluginProperties.debug()) {
                System.out.println("************ INITIALIZING " + p.getName()
                                   + " ********");
            }
            p.init();
            serviceAdded(p.getProperties().getProvisions());
        }
    }

    /**
     * Return the ClassLoader instance used to load the plugins.
     */
    public PluginClassLoader getPluginClassLoader() {
        return _pluginClassLoader;
    }

    /**
     * Returns the ClassLoader instance used to load the user defined .
     */
    public ClassLoader getBottomClassLoader() {
        return _bottomClassLoader;
    }

    /**
     * Returns the system classloader
     */
    public ClassLoader getSystemClassLoader() {
        return _systemClassLoader;
    }

    /**
     * Get the PluginLoader used to create plugin instances
     */
    PluginLoaderComposition getPluginLoader() {
        return _loader;
    }

    /**
     * Finds all candidate locations and tries to load a plugin from each
     * location. The currently configured <code>PluginLoader</code> is
     * used for the job.
     **/
    public synchronized void loadPlugins() {
        Collection all = _loader.loadPlugins();
    }

    /**
     * Loads a plugin from the given URL, if possible. The currently
     * configured <code>PluginLoader</code> is used for the job.
     *
     * @param url  the plugin's location as <code>URL</code>.
     **/
    public synchronized IPlugin loadPlugin(URL url) {
        return _loader.loadPluginFromURL(url);
    }

    /**
     * Return a list of all loaded Plugins.
     */
    public List getPlugins() {
        return _dependencyList.getFulfilledObjects();
    }

    /**
     * Register the given command. It will be called when the user
     * types the given String into the console.
     * It is not checked whether the String already exists,
     * so an existing command can be overridden!
     *
     * @param command The String identifying
     * @param clcommand The command to add
     */
    public void addCLCommand(String command, CLCommand clcommand) {
        _commands.put(command, clcommand);
    }

    /**
     * Unregister the command identified by the given String.
     */
    public void removeCLCommand(String command) {
        _commands.remove(command);
    }

    /**
     * Return a Map containing all commands as values,
     * with the identifying Strings as keys.
     */
    public Map getCLCommands() {
        return _commands;
    }

    /**
     * Adds the given plugin to the set of exit blockers.
     * If the exit blocker set becomes non-empty, the plugin system will
     * not be automatically terminated.
     *
     * @param blocker the plugin that prevents automatic termination of the
     *                plugin system.
     **/
    public void blockExit(IPlugin blocker) {
        if (blocker != null) {
            if (PluginProperties.debug()) {
                System.out.println("PluginManager: registering exit blocker "
                                   + blocker);
            }
            _blockers.add(blocker);
        }
    }

    /**
     * Removes the given plugin from the set of exit blockers.
     * If the exit blocker set becomes empty, the plugin system is
     * automatically terminated.
     *
     * @param blocker the plugin that should not any longer prevent
     *                automatic termination of the plugin system.
     *                If this plugin object is not included in the set of
     *                exit blockers, the set will not be modified.
     *                Nevertheless, the set will be checked for emptyness.
     **/
    public void exitOk(IPlugin blocker) {
        synchronized (_blockers) {
            if (blocker != null) {
                if (PluginProperties.debug()) {
                    System.out.println("PluginManager: unregistering exit blocker "
                                       + blocker);
                }
                _blockers.remove(blocker);
            }
            checkExit();
        }
    }

    /**
     * Checks whether there are any exit blockers registered, and shuts
     * down the plugin system, if not.
     **/
    private void checkExit() {
        synchronized (_blockers) {
            if (_blockers.isEmpty()) {
                System.out.println("PluginManager: no active plugins, shutting down.");

                // Since there might still be AWT threads running,
                // shut down the Java VM manually.
                // The asynchronous stop method is used to decouple the system
                // termination from the plugin that unknowingly initiated the
                // process. This also unrolls the endless loop when a plugin on
                // termination deregisters itself as exit blocker.
                stop();
            }
        }
    }

    /**
     * Terminates all plugins and the whole Java system. Whenever the
     * termination of any plugin fails, the termination process is
     * cancelled, too.
     * <p>
     * This method works asynchronously, it returns to the caller
     * immediately after initiating the termination process.
     * Plugins are then terminated in sequence (with respect to plugin
     * dependencies), but concurrently to the calling thread.
     * Multiple concurrent calls to this method will be sequentialized
     * because each termination process is synchronized on the PluginManager
     * object.
     * </p>
     * There is no feedback to the calling thread, instead messages will be
     * printed to <code>System.out</code>.
     **/
    public void stop() {
        if (PluginProperties.debug()) {
            System.out.println("Initiating PluginManager termination.");
        }
        Thread terminationThread = new Thread() {
            public void run() {
                if (stopSynchronized()) {
                    System.exit(0);
                }
            }
        };
        terminationThread.start();
    }

    /**
     * Stops the given plugin.
     * <p>
     * This method works asynchronously, it returns to the caller
     * immediately after initiating the termination process.
     * The plugin is then terminated (with respect to plugin dependencies)
     * concurrently to the calling thread.
     * Multiple concurrent calls to this method will be sequentialized
     * because each termination process is synchronized on the PluginManager
     * object.
     * </p>
     * There is no feedback to the calling thread, instead messages will be
     * printed to <code>System.out</code>.
     *
     * @param p the plugin to remove from the plugin system.
     **/
    public void stop(final IPlugin p) {
        if (PluginProperties.debug()) {
            System.out.println("Initiating termination of " + p + ".");
        }
        Thread terminationThread = new Thread() {
            public void run() {
                stopSynchronized(p);
            }
        };
        terminationThread.start();
    }

    /**
     * Stops the given plugins. Works through the list <em>from back to
     * front</em>. If any plugin cannot be stopped due to some other dependent
     * plugin (it does not matter whether it is or is not included in the
     * list in front of the plugin it depends on), the list is not further
     * processed.
     * <p>
     * Example: Assume that plugin B depends on plugin A.
     * A call with the list <code>[A, B]</code> will succeed.
     * But a call with the list <code>[B, A]</code> will fail because
     * plugin A cannot be terminated due to the dependent plugin B.
     * </p>
     * <p>
     * The reason for the backward processing is that lists of recursively
     * dependent plugins are easier to generate with the most basic plugin
     * at first position.
     * </p>
     * <p>
     * This method works asynchronously, it returns to the caller
     * immediately after initiating the termination process.
     * The plugins are then terminated (with respect to plugin dependencies)
     * concurrently to the calling thread.
     * Multiple concurrent calls to this method will be sequentialized
     * because each termination process is synchronized on the PluginManager
     * object.
     * </p>
     * There is no feedback to the calling thread, instead messages will be
     * printed to <code>System.out</code>.
     *
     * @param plugins the plugins to remove from the plugin system.
     **/
    public void stop(final List plugins) {
        if (PluginProperties.debug()) {
            System.out.println("Initiating termination of " + plugins + ".");
        }
        Thread terminationThread = new Thread() {
            public void run() {
                stopSynchronized(plugins);
            }
        };
        terminationThread.start();
    }

    /**
     * Tells whether the plugin system is about to terminate.
     * This method only returns true, if the {@link #stop()}
     * method has been used to trigger the system termination.
     * <p>
     * Please be aware that this method's result may be out of
     * date before you are able to interpret it. It is intended
     * to be used by terminated plugins that want to know
     * whether it's just them or the whole system...
     * </p>
     *
     * @return <code>true</code> if the concurrent system
     *         termination thread is running.
     **/
    public boolean isStopping() {
        return isTerminating;
    }

    /**
     * Terminates all plugins.
     * This method is called internally by {@link stop()}.
     **/
    private synchronized boolean stopSynchronized() {
        if (PluginProperties.debug()) {
            System.out.println("Stopping plugin system asynchronously.");
        }
        isTerminating = true;
        try {
            List pls = _dependencyList.getFulfilledObjects();
            return stopSynchronized(pls);
        } finally {
            isTerminating = false;
        }
    }

    /**
     * Terminates all plugins in the given list.
     * This method is called internally by {@link stopSynchronized()}
     * and {@link stop(List)}.
     **/
    private synchronized boolean stopSynchronized(List pls) {
        int limit = pls.size();
        for (int i = pls.size() - 1; i >= 0; i--) {
            // ask all plugins if they can be stopped.
            final IPlugin toStop = (IPlugin) pls.get(i);
            Object sync = new Object();
            int waitSeconds = toStop.getProperties().getIntProperty("cleanupTimeout",
                                                                    CLEANUP_TIMEOUT);
            int firstWait = waitSeconds;
            int secondWait = 1;
            if (waitSeconds > 10) {
                firstWait = waitSeconds / 10;
                secondWait = waitSeconds - firstWait;
            }
            if (PluginProperties.debug()) {
                System.out.println("waiting for " + waitSeconds + " seconds.");
            }
            SynchronizedThread ct = new SynchronizedThread(new Command() {
                    public boolean execute() {
                        return toStop.canShutDown();
                    }
                }, sync);
            try {
                ct.start();
                synchronized (sync) {
                    // stay a while...
                    if (waitSeconds == -1) {
                        // stay... forever!
                        sync.wait();
                    } else {
                        sync.wait(firstWait);
                        if (!ct.didFinish()) {
                            System.err.println("waiting for " + toStop
                                               + " to confirm termination...");
                            sync.wait(secondWait);
                        }
                    }
                }
            } catch (InterruptedException e) {
            }
            if (!ct.hadSuccess()) {
                System.err.println(toStop
                                   + " did not confirm termination in time.");
                return false;
            } else if (!ct.hadSuccess()) {
                System.err.println(toStop + " said it cannot shut down.");
                return false;
            } else {
                if (PluginProperties.debug()) {
                    System.out.println(toStop + " said it can shut down.");
                }
            }
        }
        for (int i = pls.size() - 1; i >= 0; i--) {
            IPlugin toStop = (IPlugin) pls.get(i);
            try {
                _dependencyList.removeElement(toStop);
            } catch (DependencyNotFulfilledException e) {
                System.out.println(e.getMessage());
                System.out.println("list of dependent plugins:");
                System.out.println(CollectionLister.toString(e.getElements()));
                return false;
            }

            if (!stopSynchronized(toStop)) {
                System.out.println("stop cancelled by plugin " + toStop);
                _dependencyList.addElement(new DependencyElement(toStop));
                return false;
            }
        }
        return true;
    }

    /**
     * Stops the given plugin.
     * This method is called internally by {@link stop(IPlugin)} and
     * {@link stopSynchronized(List)}.
     **/
    private synchronized boolean stopSynchronized(final IPlugin p) {
        try {
            synchronized (_dependencyList) {
                _dependencyList.removeElement(p);
            }
        } catch (DependencyNotFulfilledException e) {
            System.out.println(e.getMessage());
            System.out.println("list of dependent plugins:");
            System.out.println(CollectionLister.toString(e.getElements()));
            return false;
        }

        if (PluginProperties.debug()) {
            System.out.println("stopping " + p);
        }
        boolean result;
        Object sync = new Object();
        int waitSeconds = p.getProperties().getIntProperty("cleanupTimeout",
                                                           CLEANUP_TIMEOUT);
        SynchronizedThread ct = new SynchronizedThread(new Command() {
                public boolean execute() {
                    return p.cleanup();
                }
            }, sync);


        // prevent stop from being called by the manager because
        // it just shut down the last plugin with an exit block.
        IPlugin blockLock = new PluginAdapter(PluginProperties
                                              .getUserProperties());
        _blockers.add(blockLock);
        try {
            ct.start();
            synchronized (sync) {
                if (!ct.didFinish()) {
                    // stay a while...
                    if (waitSeconds == -1) {
                        // stay... forever!
                        sync.wait();
                    } else {
                        sync.wait(waitSeconds);
                    }
                }
            }
        } catch (InterruptedException e) {
            if (PluginProperties.debug()) {
                System.out.println("stop of " + p + " was interrupted: "
                                   + e.getMessage());
            }
        }
        _blockers.remove(blockLock);
        if (PluginProperties.debug()) {
            System.out.println(p + " stopped");
        }
        result = ct.hadSuccess();
        if (!result) {
            synchronized (_dependencyList) {
                _dependencyList.addElement(new DependencyElement(p));
            }
        } else {
            //success:
            serviceRemoved(p.getProperties().getProvisions());
        }
        return result;
    }

    /**
     * Start the Renew Plugin System.
     */
    public static void main(String[] args) {
        if (PluginProperties.debug()) {
            // show Renew splash screen
            String[] lines = new String[6];
            lines[0] = "***** ";
            lines[1] = "*    *  **  * **   **  *   *";
            lines[2] = "*****  *  * **  * *  * *   *";
            lines[3] = "*  *   **** *   * **** * * *";
            lines[4] = "*   *  *    *   * *    * * *";
            lines[5] = "*   *   *** *   *  ***  * *";

            for (int i = 0; i < lines.length; i++) {
                System.out.println(lines[i]);
            }
        }

        CLCommand command = null;
        PluginManager pm = getInstance();
        pm.loadPlugins();
        Iterator plugins = pm.getPlugins().iterator();
        while (plugins.hasNext()) {
            System.out.println("loaded plugin: "
                               + ((IPlugin) plugins.next()).getName());
        }
        if (args.length > 0) {
            String[] rest = new String[args.length - 1];
            for (int i = 0; i < args.length - 1; i++) {
                rest[i] = args[i + 1];
            }
            command = (CLCommand) pm._commands.get(args[0]);
            if (command != null) {
                if (PluginProperties.debug()) {
                    StringBuffer fullCommandLine = new StringBuffer();
                    fullCommandLine.append(args[0]);
                    for (int i = 0; i < rest.length; i++) {
                        fullCommandLine.append(' ');
                        fullCommandLine.append(rest[i]);
                    }
                    System.out.println("Executing initial command: "
                                       + fullCommandLine.toString());
                }
                command.execute(rest);
            } else {
                System.err.println("WARNING: Unknown initial command (ignored): "
                                   + args[0]);
            }
        }
        pm.checkExit();
    }

    private void initLocationFinders() {
        URL url = getClass().getProtectionDomain().getCodeSource().getLocation();
        try {
            url = new URL(url, "plugins/");
        } catch (MalformedURLException e) {
            System.err.println("Could not deduce plugins directory near plugin loader: "
                               + e);
            try {
                url = new URL(new File(System.getProperty("user.dir")).toURL(),
                              "plugins/");
            } catch (MalformedURLException e2) {
                System.err.println("Could not deduce plugins directory near current directory: "
                                   + e2);
                return;
            }
        }
        _locationFinder.addLocationFinder(new PluginSubDirFinder(url));
        _locationFinder.addLocationFinder(new PluginJarLocationFinder(url));
        Iterator fromFile = getLocations();
        while (fromFile.hasNext()) {
            url = (URL) fromFile.next();
            _locationFinder.addLocationFinder(new PluginSubDirFinder(url));
            _locationFinder.addLocationFinder(new PluginJarLocationFinder(url));
        }
    }

    /*
     * return a list of URLs to be added to the location finder
     */
    private Iterator getLocations() {
        Vector result = new Vector();
        Properties userProps = PluginProperties.getUserProperties();
        if (userProps.getProperty("pluginLocations") == null) {
            System.out.println("no additional plugin locations set.");
        }
        Collection locations = PropertyHelper.parseListString(userProps
                                                              .getProperty("pluginLocations",
                                                                           ""));
        Iterator files = locations.iterator();
        while (files.hasNext()) {
            String file = (String) files.next();
            if (PluginProperties.debug()) {
                System.out.println("location: " + file);
            }
            URL url = createURLfromString(file);
            if (PluginProperties.debug()) {
                System.out.println(url);
            }
            result.add(url);
        }
        return result.iterator();
    }

    private URL createURLfromString(String str) {
        URL url = null;

        // first check if str is a filename of an existing file
        File file = new File(str);
        if (file.exists()) {
            try {
                url = file.toURI().toURL();
                return url;
            } catch (MalformedURLException e) {
                System.out.println(e);
            }
        }

        // then, try to create a url brute force
        try {
            url = new URL(str);
        } catch (MalformedURLException e) {
        }
        return url;
    }

    /**
     * This command prints a list of all available commands.
     */
    public class HelpCommand implements CLCommand {
        /*
        * print all available commands
        */
        public void execute(String[] args) {
            System.out.println("available commands:");
            Iterator keys = _commands.keySet().iterator();
            while (keys.hasNext()) {
                String key = (String) keys.next();
                CLCommand command = (CLCommand) _commands.get(key);
                System.out.println(key + "\t-\t" + command.getDescription());
            }
        }

        public String getDescription() {
            return "print a list of all commands";
        }
    }

    /**
     * This class represents a timeout when trying to clean up
     * a plugin.
     * An object must be given to synchronize on; this will
     * be notified when the cleanup worked.
     */
    private static class SynchronizedThread extends Thread {
        private boolean _success = false;
        private boolean _finished = false;
        private Object _toNotify;
        private Command _toExecute;

        public SynchronizedThread(Command toExecute, Object toNotify) {
            _toExecute = toExecute;
            _toNotify = toNotify;
        }

        public void run() {
            //            System.out.println("cleaning up");
            _success = _toExecute.execute();
            synchronized (_toNotify) {
                //                System.out.println("ok, notifying.");
                _finished = true;
                _toNotify.notify();
            }
        }

        public boolean hadSuccess() {
            return _success;
        }

        public boolean didFinish() {
            return _finished;
        }
    }

    private static interface Command {
        public boolean execute();
    }
}