package de.renew.plugin.command;

import de.renew.plugin.CollectionLister;
import de.renew.plugin.IPlugin;
import de.renew.plugin.PluginManager;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.Vector;


/**
 * This command unloads the plugin with the given name.
 * If the argument "-v" is given, verbose output is printed,
 * especially concerning load errors regarding dependency problems.
 *
 * @author J&ouml;rn Schumacher
 */
public class UnloadCommand implements CLCommand {
    public void execute(String[] args) {
        boolean verbose = false;
        boolean recursive = false;
        Vector ignoreIndices = new Vector();
        for (int i = 0; i < args.length; i++) {
            if (args[i].equals("-v")) {
                verbose = true;
                ignoreIndices.add(new Integer(i));
            } else if (args[i].equals("-r")) {
                recursive = true;
                ignoreIndices.add(new Integer(i));
            }
        }
        String pluginName = mergeArgs(args, ignoreIndices);

        IPlugin toShutDown = PluginManager.getInstance().getPluginByName(pluginName);
        if (toShutDown == null) {
            System.out.println("UnloadCommand: did not find plugin with name "
                               + pluginName);
            return;
        }

        try {
            Map deps = checkDependencies(toShutDown);
            Set entries = deps.entrySet();
            if (entries.isEmpty()) {
                PluginManager.getInstance().stop(toShutDown);
            } else if (recursive) {
                List dependers = getDependers(toShutDown);
                System.out.println("recursively unloading "
                                   + CollectionLister.toString(dependers));
                PluginManager.getInstance().stop(dependers);
            } else {
                System.out.println("UnloadCommand: could not unload "
                                   + toShutDown
                                   + ": there are plugins depending on it.");
                if (verbose) {
                    Iterator entryIt = entries.iterator();
                    while (entryIt.hasNext()) {
                        Entry ent = (Entry) entryIt.next();
                        System.out.print(ent.getKey() + ":\t");
                        System.out.println(CollectionLister.toString((Collection) ent
                                                                     .getValue()));
                    }
                } else {
                    System.out.println("Use \"unload -v\" to see a list.");
                }
            }
        } catch (ArrayIndexOutOfBoundsException e) {
        } catch (NumberFormatException e) {
            System.out.println("cleanup canceled.");
        } catch (Exception e) {
            System.out.println("cleanup canceled: " + e + "; " + e.getMessage());
            e.printStackTrace();
        }
    }

    private String mergeArgs(String[] args, Vector ignoreIndices) {
        String result = "";
        for (int i = 0; i < args.length; i++) {
            if (!ignoreIndices.contains(new Integer(i))) {
                if (!"".equals(result)) {
                    result += " ";
                }
                result += args[i];
            }
        }
        return result;
    }

    /**
     * Check whether the given plugin may be unloaded with regards to
     * require/depend-relation.
     * Returns a list of plugins that are dependent on the given plugin.
     */
    private Map checkDependencies(IPlugin toCheck) {
        Collection provisions = toCheck.getProperties().getProvisions();
        Collection plugins = PluginManager.getInstance().getPlugins();
        HashMap dependers = new HashMap();
        Iterator prIt = provisions.iterator();
        while (prIt.hasNext()) {
            String prov = (String) prIt.next();
            Collection reqs = getRequirers(prov);
            if (!reqs.isEmpty()) {
                dependers.put(prov, reqs);
            }
        }
        return dependers;
    }

    /*
     * These two methods call themselves, recursively,
     * to find all plugins that are
     */
    private List getDependers(IPlugin toCheck) {
        List result = new Vector();
        result.add(toCheck);
        Map dependers = checkDependencies(toCheck);
        Collection deps = dependers.values();
        result.addAll(getDependers(deps));
        return result;
    }

    private List getDependers(Collection plugins) {
        List result = new Vector();
        Iterator it = plugins.iterator();
        while (it.hasNext()) {
            Object o = it.next();
            List toAdd;
            if (o instanceof Collection) {
                Collection toCheck = (Collection) o;
                toCheck.removeAll(result);
                toAdd = getDependers(toCheck);
            } else {
                IPlugin toCheck = (IPlugin) o;
                toAdd = getDependers(toCheck);
            }

            // make sure the depending plugins are not added more than once
            result.removeAll(toAdd);
            result.addAll(toAdd);
        }
        return result;
    }

    private Collection getRequirers(String prov) {
        Collection result = new Vector();
        Collection plugins = PluginManager.getInstance().getPlugins();

        Iterator plIt = plugins.iterator();

        // collect all plugins that require the current provision
        while (plIt.hasNext()) {
            IPlugin pl = (IPlugin) plIt.next();
            Collection reqs = pl.getProperties().getRequirements();
            if (reqs.contains(prov)) {
                result.add(pl);
            }
        }
        return result;
    }

    public String getDescription() {
        return "cleans up (deactivates) a plugin";
    }
}