package de.renew.plugin.load;

import de.renew.plugin.CollectionLister;
import de.renew.plugin.DependencyCheckList;
import de.renew.plugin.DependencyCheckList.DependencyElement;
import de.renew.plugin.DependencyNotFulfilledException;
import de.renew.plugin.IPlugin;
import de.renew.plugin.PluginManager;
import de.renew.plugin.PluginProperties;
import de.renew.plugin.locate.PluginLocationFinders;

import java.net.URL;

import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;


/**
 * This Class is a Composition of PluginLoaders.
 * When an instance's loadPlugins() method is called, all contained
 * PluginLoaders are triggered.
 * <p>
 * In the current implementation, this class checks all dependencies and
 * decides when to add which plugin to the {@link PluginManager}. So this
 * class <i>must</i> be used as top level plugin loader, and it <i>may
 * not</i> be added as a tree node to the plugin loader hierarchy.
 * However, the implementation does not check for these conditions.
 * In future versions, dependency check and plugin addition can be moved
 * somewhere else, where the functions are more appropriate.
 * </p>
 * @author Joern Schumacher
 * @author Michael Duvigneau
 */
public class PluginLoaderComposition implements PluginLoader {
    private Collection _loaders = new Vector();

    /**
     * Constructor for PluginLoaderComposition.
     */
    public PluginLoaderComposition() {
        super();
    }

    public Collection loadPlugins() {
        System.out.println("loading plugins...");
        DependencyCheckList dependencyList = new DependencyCheckList();

        Collection locations = PluginLocationFinders.getInstance()
                                                    .findPluginLocations();
        Collection result = new Vector();


        Iterator all = locations.iterator();
        while (all.hasNext()) {
            // System.out.println (all.next().getClass());
            PluginProperties props = (PluginProperties) all.next();
            dependencyList.addElement(new DependencyElement(props));
        }

        List fulfilledDependencies = dependencyList.getFulfilledObjects();
        if (PluginProperties.debug()) {
            Iterator it = fulfilledDependencies.iterator();
            while (it.hasNext()) {
                System.out.println(it.next());
            }
        }
        for (int i = 0; i < fulfilledDependencies.size(); i++) {
            PluginProperties props = (PluginProperties) fulfilledDependencies
                                     .get(i);
            if (PluginManager.getInstance().getPluginByName(props.getName()) != null) {
                System.out.println("PluginLoader: " + props.getName()
                                   + " has already been loaded.");
                // do not load if this plugin has already been loaded
                continue;
            }
            URL url = props.getURL();
            Iterator it = _loaders.iterator();
            IPlugin loaded = null;
            while (it.hasNext()) {
                try {
                    PluginLoader loader = (PluginLoader) it.next();
                    loaded = loader.loadPlugin(props);
                    if (loaded != null) {
                        break;
                    }
                } catch (RuntimeException e) {
                    System.err.println("failure to load this plugin: " + e);
                    if (PluginProperties.debug()) {
                        e.printStackTrace();
                    }
                } catch (NoClassDefFoundError e) {
                    System.err.println("failure to load this plugin: " + e);
                    if (PluginProperties.debug()) {
                        e.printStackTrace();
                    }
                }
            }
            if (loaded != null) {
                try {
                    PluginManager.getInstance().addPlugin(loaded);
                    result.add(loaded);
                } catch (DependencyNotFulfilledException e) {
                    assert false : "The PluginManager doubts our dependency check for "
                    + loaded;
                }
            } else {
                if (PluginProperties.debug()) {
                    System.out.println("PluginLoaderComposition: did not load plugin from "
                                       + props);
                }
                Collection retracted = dependencyList
                                       .removeElementWithDependencies(props);
                if (!retracted.isEmpty()) {
                    if (PluginProperties.debug()) {
                        System.out.println("PluginLoaderComposition: recalculated dependencies.");
                    }
                }

                // There is an important (dirty?) assumption here:
                // Because the dependencyList is ordered, the
                // fulfilledDependencies list should not differ in the
                // first part before the failed plugin. So we can just
                // continue to loop from the same point where we are.
                fulfilledDependencies = dependencyList.getFulfilledObjects();
                i--;
            }
        }

        Collection unfulfilled = dependencyList.getUnfulfilled();
        if (!unfulfilled.isEmpty()) {
            System.err.println("WARNING: There are plugins with unfulfilled dependencies:");
            System.err.println(CollectionLister.toString(unfulfilled));
        }

        return result;
    }

    public void addLoader(PluginLoader l) {
        _loaders.add(l);
    }

    public void removeLoader(PluginLoader l) {
        _loaders.remove(l);
    }

    /**
     * This method is called to trigger the loading of a plugin with
     * the previously loaded properties..
     * All previously registered loaders are called to try to load
     * a plugin; the first successfully loaded is returned.
     * <p>
     * In contrast to {@link #loadPluginFromURL} and {@link loadPlugins},
     * this method does neither check dependencies nor add the plugin to
     * the <code>PluginManager</code> automatically.
     * </p>
     */
    public IPlugin loadPlugin(PluginProperties props) {
        IPlugin result = null;
        Iterator i = _loaders.iterator();
        PluginLoader loader = null;
        while (i.hasNext()) {
            try {
                loader = (PluginLoader) i.next();
                result = loader.loadPlugin(props);
                if (result != null) {
                    break;
                }
            } catch (Exception e) {
                System.err.println("PluginLoaderComposition: " + e
                                   + " occured when " + loader
                                   + " tried to load plugin from "
                                   + props.getURL());
                e.printStackTrace();
            }
        }
        return result;
    }

    /**
     * This method is called to trigger the loading of a plugin at
     * the given location.
     * <p>
     * There are two ways of loading the plugin, they are tried in the
     * given order until a plugin has been loaded.
     * <ol>
     * <li><code>PluginLocationFinders</code> are asked to transform the
     *     given <code>source</code> URL into <code>PluginProperties</code>.
     *     If the finders are successful, the plugin is loaded according to
     *     the properties found.
     * </li>
     * <li>All registered <code>PluginLoader</code>s are called to try to
     *     load the plugin without the assistance of a finder; the first
     *     successfully loaded plugin is returned.
     * </li>
     * </ol>
     * If a plugin has been loaded, it is automatically added to the
     * <code>PluginManager</code>.
     * </p>
     **/
    public IPlugin loadPluginFromURL(URL source) {
        IPlugin result = null;
        PluginManager mgr = PluginManager.getInstance();
        PluginProperties props = PluginLocationFinders.getInstance()
                                                      .checkPluginLocation(source);
        if (props != null) {
            if (mgr.checkDependenciesFulfilled(props)) {
                result = loadPlugin(props);
            } else {
                System.err.println("Dependencies are not fulfilled for plugin "
                                   + props.getName() + " located at "
                                   + props.getURL());
            }
        } else {
            Iterator i = _loaders.iterator();
            PluginLoader loader = null;
            while (i.hasNext()) {
                try {
                    loader = (PluginLoader) i.next();
                    result = loader.loadPluginFromURL(source);
                    if (result != null) {
                        break;
                    }
                } catch (Exception e) {
                    System.err.println("PluginLoaderComposition: " + e
                                       + " occured when " + loader
                                       + " tried to load plugin from " + source);
                    if (PluginProperties.debug()) {
                        e.printStackTrace();
                    }
                }
            }
        }
        if (result != null) {
            try {
                mgr.addPlugin(result);
            } catch (DependencyNotFulfilledException e) {
                System.err.println("Dependencies are not fulfilled for "
                                   + result + " loaded from " + source);
            }
        }
        return result;
    }
}