/*
 * (c) Copyright 2001 MyCorporation.
 * All Rights Reserved.
 */
package de.renew.plugin.load;

import de.renew.plugin.IPlugin;
import de.renew.plugin.PluginAdapter;
import de.renew.plugin.PluginClassLoader;
import de.renew.plugin.PluginManager;
import de.renew.plugin.PluginProperties;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import java.net.MalformedURLException;
import java.net.URL;

import java.util.Collection;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Vector;


/**
 * Singleton support class that loads static plugins.
 *
 * There are two ways to instantiate a plugin with this class' loadPlugin(PluginProperties) method,
 * depending on what a call to getURL() on the given properties object returns:
 * if it is a jar file, this jar file will be assumed to contain the plugin code
 * as well as the plugin.cfg file with the plugin's configuration information;
 * if it is a directory, the directory is searched for the plugin.cfg file as well as
 * for a jarfile which is then used as the plugin code source.
 * If there is no jarfile in the directory, the directory itself is used as code source for
 * the resulting plugin.
 */
public class SimplePluginLoader implements PluginLoader {
    /*
     * Initializes the PluginLoader instance.
     */
    public SimplePluginLoader() {
    }

    /**
     * Loads the plugin in the given directory.
     * For that, the plugin.cfg file is parsed and searched for
     * the pluginClassName entry.
     * This class will then be loaded and instantiated.
     */
    public IPlugin loadPlugin(PluginProperties props) {
        URL pluginLocation = props.getURL();
        if (PluginProperties.debug()) {
            System.out.println("SimplePluginLoader loading from "
                               + props.getURL());
        }
        URL[] pluginJars = unifyURL(pluginLocation);
        PluginClassLoader pluginClassLoader = PluginManager.getInstance()
                                                           .getPluginClassLoader();
        String message = "ok.";


        IPlugin plugin = null;
        try {
            for (int x = 0; x < pluginJars.length; x++) {
                pluginClassLoader.addURL(pluginJars[x]);
            }

            String className = props.getProperty("mainClass");


            if (className != null) {
                if (PluginProperties.debug()) {
                    System.out.println("* creating a " + className
                                       + " with cl " + pluginClassLoader);
                }
                Class clazz = pluginClassLoader.loadClass(className);
                if (clazz.getClassLoader() != pluginClassLoader) {
                    System.err.println("Warning: system classloader was used to load "
                                       + clazz
                                       + "; it was probably in your classpath.");
                    System.err.println("That might turn out as a problem.");
                }

                Class[] params = { props.getClass() };
                Constructor cons = clazz.getConstructor(params);
                Object[] args = { props };
                plugin = (IPlugin) cons.newInstance(args);

                if (PluginProperties.debug()) {
                    System.out.println("SimplePluginLoader: instantiated "
                                       + plugin);
                }
            } else {
                if (PluginProperties.debug()) {
                    System.out.println("* no main class!");
                }
                plugin = new PluginAdapter(props);
            }
        } catch (InvocationTargetException ite) {
            Throwable e = ite.getTargetException();
            message = e.toString();
        } catch (Exception e) {
            message = e.toString();
        }
        if (plugin == null) {
            String desc = props.getName();
            if (desc == null) {
                desc = "from " + pluginLocation;
            }
            System.err.println("SimplePluginLoader: Could not load plugin "
                               + desc + ": " + message);
        } else if (PluginProperties.debug()) {
            System.out.println("SimplePluginLoader result: " + message);
        }
        return plugin;
    }

    public IPlugin loadPluginFromURL(URL url) {
        PluginProperties props = new PluginProperties(url);
        PluginConfigFinder finder = new PluginConfigFinder();
        try {
            InputStream stream = PluginConfigFinder.getConfigInputStream(url);
            props.load(stream);
            return loadPlugin(props);
        } catch (MalformedURLException e) {
            System.out.println("SimplePluginLoader.loadPluginFromURL: " + e);
        } catch (IOException e) {
            System.out.println("SimplePluginLoader.loadPluginFromURL: " + e);
        }
        return null;
    }

    private URL[] getURLsfromDirectory(File location) {
        // changed (06.05.2004) to support more than one jar in a plugin folder
        if (location.getName().endsWith("jar")) {
            try {
                return new URL[] { location.toURI().toURL() };
            } catch (MalformedURLException e) {
                System.err.println("SimplePluginLoader: Could not convert to URL: "
                                   + location + " (" + e.getMessage() + ").");
            }
        }
        if (!location.isDirectory()) {
            location = location.getParentFile();
        }
        File[] jarFiles = location.listFiles(new FilenameFilter() {
                public boolean accept(File dir, String file) {
                    return file.endsWith(".jar");
                }
            });

        List locs = new Vector();
        if (jarFiles.length > 0) {
            for (int x = 0; x < jarFiles.length; x++) {
                File jarFile = jarFiles[x];
                try {
                    locs.add(jarFile.toURL());
                } catch (MalformedURLException e) {
                    System.out.println("PluginLoader: could not create URL from "
                                       + jarFile);
                }
            }
        } else {
            System.err.println("SimplePluginLoader: no jarfile found in "
                               + location + "; using dir.");
            try {
                locs.add(location.toURI().toURL());
            } catch (MalformedURLException e) {
                System.err.println("SimplePluginLoader: could not convert dir to URL: "
                                   + location + " (" + e.getMessage() + ").");
            }
        }
        if (locs.size() > 0) {
            return (URL[]) locs.toArray(new URL[] {  });
        }
        return null;
    }

    private URL[] unifyURL(URL given) {
        String path = given.getFile();
        File dir = new File(path);
        URL[] jarURLs = getURLsfromDirectory(dir);
        if (jarURLs == null) {
            System.out.println("no jar!");
            return new URL[] { given };
        }
        return jarURLs;
    }

    /**
     * Converts a String into a list by separating by commas
     */
    protected Collection parseListString(String list) {
        StringTokenizer tok = new StringTokenizer(list, ",");
        Collection result = new Vector(tok.countTokens());
        try {
            for (int i = 0; tok.hasMoreTokens(); i++) {
                String currentToken = tok.nextToken();
                result.add(currentToken);
            }
        } catch (ArrayIndexOutOfBoundsException e) {
            System.err.println("PluginLoader: " + e + " when parsing " + list
                               + " as list!");
        }
        return result;
    }

    public static class PluginConfigFinder {
        public static InputStream getConfigInputStream(URL pluginLocation)
                throws MalformedURLException, IOException {
            URL configLocation;
            configLocation = new URL(pluginLocation, "plugin.cfg");
            //		System.out.println ("loading cfg from " + configLocation.toExternalForm());
            InputStream stream = null;
            try {
                stream = configLocation.openStream();
            } catch (Exception e) {
                // maybe its a jar. try on.
                String urlString = pluginLocation.toExternalForm();
                int index = urlString.lastIndexOf('!');
                if (index > -1) {
                    urlString = urlString.substring(0, index) + "/plugin.cfg";
                    if (PluginProperties.debug()) {
                        System.out.println("trying alternate URL for plugin.cfg: "
                                           + urlString);
                    }
                }
                configLocation = new URL(urlString);
                stream = configLocation.openStream();
            }
            return stream;
        }
    }

    public static class PluginSourceFinder {
        public static URL getPluginSource(URL pluginLocation) {
            return pluginLocation;
        }
    }
}