package de.renew.engine.simulator;

import de.renew.engine.searcher.SilentTracerFactory;
import de.renew.engine.searcher.TracerFactory;


public class ParallelSimulator implements Simulator {
    public final int multiplicity;
    private final Simulator[] simulators;
    private final TracerFactory tracerFactory;

    public ParallelSimulator(TracerFactory tracerFactory, int multiplicity,
                             boolean wantEventQueueDelay) {
        if (tracerFactory == null) {
            tracerFactory = new SilentTracerFactory();
        }
        this.tracerFactory = tracerFactory;

        if (multiplicity == 0) {
            multiplicity = 1;
        }
        boolean wantConcurrent = multiplicity > 0;
        if (!wantConcurrent) {
            multiplicity = -multiplicity;
        }
        this.multiplicity = multiplicity;

        simulators = new Simulator[multiplicity];
        for (int i = 0; i < multiplicity; i++) {
            TracerFactory childTracerFactory = tracerFactory.newChildFactory(i
                                                                             + "-");
            if (wantConcurrent) {
                simulators[i] = new ConcurrentSimulator(childTracerFactory,
                                                        wantEventQueueDelay);
            } else {
                simulators[i] = new NonConcurrentSimulator(childTracerFactory,
                                                           wantEventQueueDelay);
            }
        }
    }

    public boolean isActive() {
        for (int i = 0; i < multiplicity; i++) {
            if (simulators[i].isActive()) {
                return true;
            }
        }
        return false;
    }

    // Start the simulation in the background.
    public synchronized void startRun() {
        for (int i = 0; i < multiplicity; i++) {
            simulators[i].startRun();
        }
    }

    // Gently stop the simulation.
    public void stopRun() {
        for (int i = 0; i < multiplicity; i++) {
            simulators[i].stopRun();
        }
    }

    // Terminate the simulation once and for all.
    // Do some final clean-up and exit all threads.
    public synchronized void terminateRun() {
        for (int i = 0; i < multiplicity; i++) {
            simulators[i].terminateRun();
        }
    }

    // Perform just one step or terminate the simulation, if
    // it is running in the background. Return true, if
    // another binding could be found.
    public synchronized int step() {
        // Stop all simulators except the first.
        for (int i = 0; i < multiplicity; i++) {
            simulators[i].stopRun();
        }


        // Try to fire any simulator and stop on the first success.
        // This is required when aggregating sequential simulators,
        // because their enabledness status might be different.
        // (Sequential simulators hold the enabled transition instance
        // hostage until the firing is over.)
        int status = 0; // will be overwritten
        for (int i = 0; i < multiplicity; i++) {
            status = simulators[i].step();
            if (status == statusStepComplete || status == statusLastComplete) {
                // One simulator performed a step and all simulators
                // are in stopped state.
                return status;
            }
        }

        // No simulator was able to perform a step.
        return status;
    }

    public void refresh() {
        for (int i = 0; i < multiplicity; i++) {
            simulators[i].refresh();
        }
    }

    /**
     * @return <code>false</code>
     **/
    public boolean isSequential() {
        return false;
    }
}