package de.renew.net;

import de.renew.engine.searcher.SilentTracer;
import de.renew.engine.searcher.Tracer;

import de.renew.plugin.PluginProperties;

import de.renew.unify.Impossible;

import de.renew.util.RenewObjectInputStream;
import de.renew.util.RenewObjectOutputStream;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.Serializable;
import java.io.StreamCorruptedException;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;


/**
 * A net, consisting of places and transitions.
 * <p>
 * Nets are not roibust against concurrent updates.
 * Specifically, it is not allowed to change the structure
 * of a net while a simulation is running. The standard
 * pattern is to create a net, make it public, and then
 * use it without ever modifying it again.
 */
public class Net implements Serializable {
    static final long serialVersionUID = 4424943946669800287L;

    /**
     * A map from names to nets. Synchronize on this map
     * to get a consistent view on the set of registered nets.
     */
    private static java.util.Map netsByName = new HashMap();
    private static NetLoader netLoader = null;
    Set places = new HashSet();
    Set transitions = new HashSet();
    String name = null;
    boolean earlyTokens = false;
    private int netCount = 0;

    public Net() {
        name = null;
    }

    /**
     * Creates an empty net body with the given name. Since the net
     * is not functional yet, it is not announced to the map for
     * the static {@link #forName} method. This has to be done
     * separately after full construction of the net by using the
     * {@link #makeKnown} method.
     **/
    public Net(String name) {
        this.name = name;
    }

    /**
     * Returns the net object with the given name.
     * If a net is not known, asks the net loader (see
     * {@link #setNetLoader}) to load the net.
     *
     * @param name  the name of the net to find.
     *
     * @return
     *   the net object with the given name.
     *   Never returns <code>null</code>.
     *
     * @throws NetNotFoundException
     *   if the net is not loaded and no loader was set or the
     *   loader could not find the net.
     **/
    public static Net forName(String name) throws NetNotFoundException {
        synchronized (netsByName) {
            Net net = (Net) netsByName.get(name);
            if (net == null) {
                if (netLoader == null) {
                    throw new NetNotFoundException(name);
                } else {
                    // Check again if the net is still unknown.
                    // Things may have changed during the synchronization sleep...
                    net = (Net) netsByName.get(name);
                    if (net == null) {
                        net = netLoader.loadNet(name);
                    }
                }
            }
            return net;

        }
    }

    /**
     * Clears the mapping used by the {@link #forName} method.
     **/
    public static void forgetAllNets() {
        synchronized (netsByName) {
            netsByName.clear();
        }
    }

    /**
     * Sets the net loader. This request will be denied when there
     * are any known nets. You are safe if you call it immediately
     * after {@link #forgetAllNets}...
     * @throws IllegalStateException
     *   if there are known nets
     **/
    public static void setNetLoader(NetLoader loader) {
        synchronized (netsByName) {
            if (netsByName.isEmpty()) {
                netLoader = loader;
            } else {
                throw new IllegalStateException("Cannot change net loader while nets are known.");
            }
        }
    }

    /**
     * Gets the net loader.
     **/
    public static NetLoader getNetLoader() {
        return netLoader;
    }

    /**
     * Returns an enumeration of all known nets.
     **/
    public static Iterator /*Net*/ allKnownNets() {
        synchronized (netsByName) {
            return new ArrayList(netsByName.values()).iterator();
        }
    }

    /**
     * Confirms the end of the construction phase of the net. The
     * net is included in the {@link #forName} lookup.
     *
     * @since Renew 1.6
     **/
    public void makeKnown() {
        synchronized (netsByName) {
            if (name != null) {
                netsByName.put(name, this);
            }
        }
    }

    public Collection places() {
        return places;
    }

    public int placeCount() {
        return places.size();
    }

    public Collection transitions() {
        return transitions;
    }

    public int transitionCount() {
        return transitions.size();
    }

    synchronized int makeNetNumber() {
        return ++netCount;
    }

    public String toString() {
        return name;
    }

    public String getName() {
        return name;
    }

    /**
     * Created one instance of this net, but do not yet confirm
     * its creation. Used for creating instances that may have to be
     * discarded early.
     *
     * @return a newly allocated <code>NetInstance</code>
     * @exception Impossible if an initial marking expression
     *            evaluation fails.
     */
    public NetInstance makeInstance() throws Impossible {
        // Instead of creating a generic NetInstance, we might also
        // want to create a specialized NetInstance that supports
        // own methods and implements interfaces.
        return new NetInstanceImpl(this);
    }

    /**
     * Unlike the method {@link #makeInstance},
     * this publicly accessible
     * method immediately confirms the creation of the
     * generated net instance, so that external users do not need
     * to know about confirmation issues.
     *
     * @param tracer the tracer that gathers information about initial tokens
     * @return a newly allocated <code>NetInstance</code>
     */
    public NetInstance buildInstance(Tracer tracer) {
        try {
            NetInstance instance = makeInstance();
            instance.createConfirmation(tracer);
            return instance;
        } catch (Impossible e) {
            throw new RuntimeException("Could not make net instance.");
        }
    }

    /**
     * Delegate to {@link #buildInstance(Tracer)} with a tracer that
     * ignores all output.
     *
     * @return a newly allocated <code>NetInstance</code>
     */
    public NetInstance buildInstance() {
        return buildInstance(SilentTracer.INSTANCE);
    }

    void add(Place place) {
        places.add(place);
    }

    void add(Transition transition) {
        transitions.add(transition);
    }

    void remove(Place place) {
        places.remove(place);
    }

    void remove(Transition transition) {
        transitions.remove(transition);
    }

    public void setEarlyTokens(boolean earlyTokens) {
        this.earlyTokens = earlyTokens;
    }

    /**
     * Serialization method, behaves like default writeObject
     * method except using the domain trace feature, if the
     * output is a RenewObjectOutputStream.
     * @see de.renew.util.RenewObjectOutputStream
     **/
    private void writeObject(java.io.ObjectOutputStream out)
            throws IOException {
        RenewObjectOutputStream rOut = null;
        if (out instanceof RenewObjectOutputStream) {
            rOut = (RenewObjectOutputStream) out;
        }
        if (rOut != null) {
            rOut.beginDomain(this);
        }
        out.defaultWriteObject();
        if (rOut != null) {
            rOut.endDomain(this);
        }
    }

    /**
     * Deserialization method, behaves like default readObject
     * method and makes the net known to <code>Net.forName()</code>.
     **/
    private void readObject(java.io.ObjectInputStream in)
            throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        makeKnown();
    }

    /**
     * Writes all currently known <code>Net</code>s
     * to the given stream. The written information
     * describes the static part of all compiled nets.
     * <p>
     * If the given <code>ObjectOutput</code> is a <code><b>
     * de.renew.util.RenewObjectOutputStream</b></code>, its
     * feature of cutting down the recursion depth by delaying
     * the serialization of some fields will be used.
     * </p>
     *
     * Added Apr 18 2000  Michael Duvigneau
     *
     * @param output target stream (see note about RenewObjectOutputStream)
     * @see de.renew.util.RenewObjectOutputStream
     **/
    public static void saveAllNets(ObjectOutput output)
            throws IOException {
        synchronized (netsByName) {
            Enumeration saveList;
            RenewObjectOutputStream rOut = null;
            if (output instanceof RenewObjectOutputStream) {
                rOut = (RenewObjectOutputStream) output;
            }

            if (rOut != null) {
                rOut.beginDomain(Net.class);
            }


            // Save all currently known nets.
            output.writeInt(netsByName.size());
            Iterator iterator = netsByName.values().iterator();
            while (iterator.hasNext()) {
                output.writeObject(iterator.next());
            }


            // If a RenewObjectOutputStream is used, write
            // all delayed fields NOW.
            if (rOut != null) {
                rOut.writeDelayedObjects();
            }

            if (rOut != null) {
                rOut.endDomain(Net.class);
            }
        }
    }

    /**
     * Reads nets stored by <code>saveAllInstances()</code>
     * and adds the new <code>Net</code>s to the list
     * of known nets by name.
     * </p><p>
     * If the given <code>ObjectInput</code> is a <code>
     * de.renew.util.RenewObjectInputStream</code>, the
     * neccessary steps to cover delayed serialization will
     * be made.
     * </p><p>
     * The ObjectInputStream will be read using
     * <code>de.renew.util.ClassSource</code> to provide
     * its ability of reloading all user defined classes.
     * </p>
     *
     * Added Apr 18 2000  Michael Duvigneau
     *
     * @param input source stream (see notes above)
     * @see #saveAllNets
     * @see de.renew.util.ClassSource
     * @see de.renew.util.RenewObjectInputStream
     **/
    public static void loadNets(ObjectInput input)
            throws IOException, ClassNotFoundException {
        // Read all stored Nets.
        int count = input.readInt();
        Net newNet;
        try {
            for (int i = 0; i < count; i++) {
                newNet = (Net) de.renew.util.ClassSource.readObject(input);
            }
        } catch (ClassCastException e) {
            if (PluginProperties.debug()) {
                e.printStackTrace();
            }
            throw new StreamCorruptedException("Object other than Net found "
                                               + "when looking for nets: "
                                               + e.getMessage());
        }


        // If a RenewObjectInputStream is used, read
        // all delayed fields NOW.
        if (input instanceof RenewObjectInputStream) {
            ((RenewObjectInputStream) input).readDelayedObjects();
        }
    }

    // ----------------------------------------------- ID handling ----


    /**
     * Return the place of the net with the given ID.
     * Return <code>null</code>, if no such place exists.
     * <p></p>
     * Added Mon Jul 17  2000 by Michael Duvigneau
     *
     * @param id the place's ID
     * @return a place or <code>null</code>
     */
    public Place getPlaceWithID(NetElementID id) {
        for (Iterator i = places().iterator(); i.hasNext();) {
            Place place = (Place) i.next();
            if (place.getID().equals(id)) {
                return place;
            }
        }
        return null;
    }

    /**
     * Return the transition of the net with the given ID.
     * Return <code>null</code>, if no such place exists.
     * <p></p>
     * Added Mon Jul 17  2000 by Michael Duvigneau
     *
     * @param id the transition's ID
     * @return a transition or <code>null</code>
     */
    public Transition getTransitionWithID(NetElementID id) {
        for (Iterator i = transitions().iterator(); i.hasNext();) {
            Transition transition = (Transition) i.next();
            if (transition.getID().equals(id)) {
                return transition;
            }
        }
        return null;
    }
}