package de.renew.net.event;

import de.renew.engine.simulator.EventQueue;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;


/**
 * A <code>ListenerSet</code> manages a collection
 * of event listeners and is able to distribute events
 * to all of its listeners.
 * <p>
 * Synchronization is done on this object.
 * <p>
 * When an event is delivered synchronously, it is ensured that
 * the listener set is locked, so that attempted changes to the
 * set of listeners do not effect the current notifications.
 * <p>
 * This is a utility class that is not type safe. It should be
 * subclassed to get adequate functionality.
 *
 * @author <a href="mailto:kummer@informatik.uni-hamburg.de">Olaf Kummer</a>
 */
public class ListenerSet {

    /**
     * The listeners to notify if events occur.
     **/
    private transient Set listeners = new HashSet();

    synchronized void include(NetEventListener listener) {
        listeners.add(listener);
    }

    synchronized void exclude(NetEventListener listener) {
        listeners.remove(listener);
    }

    /**
     * Call the dispatcher once for each registered listener.
     **/
    synchronized void dispatch(final ListenerSetDispatcher dispatcher) {
        Iterator enum = listeners.iterator();
        while (enum.hasNext()) {
            final NetEventListener listener = (NetEventListener) enum.next();
            if (listener.wantSynchronousNotification()) {
                dispatcher.dispatchTo(listener);
            } else {
                EventQueue.enqueue(new Runnable() {
                        public void run() {
                            dispatcher.dispatchTo(listener);
                        }
                    });
            }
        }
    }
}