package de.renew.shadow;

import de.renew.net.NetNotFoundException;

import de.renew.plugin.PluginProperties;

import de.renew.util.ClassSource;
import de.renew.util.StringUtil;

import java.io.File;
import java.io.InputStream;
import java.io.ObjectInput;
import java.io.ObjectInputStream;

import java.net.URL;

import java.util.Iterator;
import java.util.Properties;


/**
 * Tries to load shadow nets from <code>.sns</code> or
 * <code>.sha</code> files. The filename (without extension) must
 * match the net name. The loader looks for these files in the
 * directories specified by the <code>de.renew.netPath</code>
 * system property.
 * <p>
 * Additionally, the shadow net system stored in the file may
 * contain only one shadow net, which must have the same name as
 * the net and the file, too. This restriction exists to avoid
 * unexpected results from loading shadow nets out of files with
 * different names.
 * </p>
 * DefaultShadowNetLoader.java
 * Created: Thu Jan 17  2002
 * @author Michael Duvigneau
 **/
public class DefaultShadowNetLoader implements ShadowNetLoader {

    /**
     * An array of path entries denoting the directories where to
     * look for serialized <code>ShadowNetSystem</code> files.
     **/
    private PathEntry[] netSource;

    /**
     * An array of file name extensions that indicate serialized
     * <code>ShadowNetSystem</code> files.
     **/
    private final String[] shadowExtensions = new String[] { ".sns", ".sha" };

    public DefaultShadowNetLoader(Properties props) {
        if (props != null) {
            configureNetPath(props);
        } else {
            configureNetPath(System.getProperties());
        }
    }

    /**
     * Configures the net search path from a given property set.
     *
     * @param props the property set to extract the
     *              <code>de.renew.netPath</code> property from.
     * @see #setNetPath(String)
     **/
    public void configureNetPath(Properties props) {
        setNetPath(props.getProperty("de.renew.netPath",
                                     System.getProperty("user.dir")));
    }

    /**
     * Sets a search path (like the CLASSPATH) to look for shadow
     * net system files when a net is missing. The directories in
     * the path are separated by {@link File#pathSeparatorChar}.
     **/
    void setNetPath(String path) {
        setNetPath(StringUtil.splitPaths(path));
    }

    /**
     * Sets search paths (like the CLASSPATH) to look for net
     * drawing files when a drawing is missing. Each String in the
     * array denotes exactly one directory to search.
     **/
    void setNetPath(String[] paths) {
        this.netSource = canonizePaths(paths);
        for (int i = 0; i < netSource.length; ++i) {
            if (PluginProperties.debug()) {
                System.out.println("Shadow net loader source"
                                   + (netSource[i].isClasspathRelative
                                      ? " (relative to CLASSPATH): " : ": ")
                                   + netSource[i].path);
            }
        }
    }

    /**
     * Creates an array of path entries from an array of path strings.
     * Strings starting with the reserved word <code>"CLASSPATH"</code> are
     * converted to classpath-relative path entries. All other path strings
     * are converted to path entries that point to canonized directory
     * names.
     *
     * @param paths  the <code>String</code> array with path names.
     * @return an array of <code>PathEntry</code> objects.
     **/
    public static PathEntry[] canonizePaths(String[] paths) {
        if (paths == null) {
            return new PathEntry[0];
        }
        PathEntry[] canonizedEntries = new PathEntry[paths.length];
        for (int i = 0; i < paths.length; ++i) {
            if (paths[i].trim().startsWith("CLASSPATH" + File.separator)) {
                canonizedEntries[i] = new PathEntry(paths[i].trim().substring(9
                                                                              + File.separator
                                                                                .length()),
                                                    true);
            } else if (paths[i].trim().equals("CLASSPATH")) {
                canonizedEntries[i] = new PathEntry("", true);
            } else {
                canonizedEntries[i] = new PathEntry(StringUtil.makeCanonical(paths[i]),
                                                    false);
            }
        }
        return canonizedEntries;
    }

    /**
     * Creates an OS-dependent path string from the given array of path
     * entries. This is the reverse function of
     * <code>paths = canonizePaths(StringUtil.splitPaths(pathString))</code>.
     *
     * @param paths an array of <code>PathEntry[]</code> objects.
     * @return a path <code>String</code> including all entries of the
     * given array. Returns <code>""</code> if the array is empty or
     * non-existant.
     **/
    public static String asPathString(PathEntry[] paths) {
        if (paths == null || paths.length == 0) {
            return "";
        }
        boolean first = true;
        StringBuffer buffer = new StringBuffer();
        for (int i = 0; i < paths.length; ++i) {
            if (paths[i] != null) {
                if (first) {
                    first = false;
                } else {
                    buffer.append(File.pathSeparator);
                }
                if (paths[i].isClasspathRelative) {
                    buffer.append("CLASSPATH");
                    if (!"".equals(paths[i].path)) {
                        buffer.append(File.separator);
                    }
                }
                buffer.append(paths[i].path);
            }
        }
        return buffer.toString();
    }

    public ShadowNet loadShadowNet(String netName, ShadowNetSystem netSystem)
            throws NetNotFoundException {
        ShadowNet loaded = findShadowFile(netName);
        if (loaded != null) {
            loaded.switchNetSystem(netSystem);
            return loaded;
        }
        throw new NetNotFoundException(netName);
    }

    /**
     * Tries to find and load a <code>.sns</code> or
     * <code>.sha</code> file for the given net name.
     * The shadow net is loaded and returned.
     * It is not checked whether a shadow net with the given name
     * is already known.
     *
     * @return the fresh shadow net, if one was loaded -
     *   <code>null</code>, if no matching file could be found.
     **/
    ShadowNet findShadowFile(String name) {
        String path;
        URL url;
        boolean error;
        int ext;
        for (int i = 0; i < netSource.length; ++i) {
            error = true;
            ext = 0;
            url = null;
            while (error && (ext < shadowExtensions.length)) {
                // First step: Generate URL from netpath, net
                // name and extension. Also check whether the
                // source file exists.
                error = false;
                StringBuffer buffer = new StringBuffer();
                buffer.append(netSource[i].path);
                if (!"".equals(netSource[i].path)) {
                    buffer.append(File.separator);
                }
                buffer.append(name);
                buffer.append(shadowExtensions[ext]);
                path = buffer.toString();
                try {
                    if (netSource[i].isClasspathRelative) {
                        url = ClassSource.getClassLoader().getResource(StringUtil
                                                                       .convertToSlashes(path));
                        error = (url == null);
                    } else {
                        File file = new File(path);
                        if (file.canRead()) {
                            url = file.toURI().toURL();
                        } else {
                            error = true;
                        }
                    }
                } catch (Exception e) {
                    System.err.println("DefaultShadowNetLoader: " + path
                                       + " caused " + e);
                    if (PluginProperties.debug()) {
                        e.printStackTrace();
                    }
                    error = true;
                }


                // Second step: If we have an URL to an existing
                // source file, read, load and return it.
                if (!error) {
                    try {
                        if (PluginProperties.debug()) {
                            System.out.println("Loading shadow net from " + url);
                        }
                        InputStream stream = url.openStream();
                        ObjectInput input = new ObjectInputStream(stream);
                        ShadowNetSystem netSystem = (ShadowNetSystem) input
                                                    .readObject();
                        input.close();

                        Iterator nets = netSystem.elements().iterator();
                        ShadowNet net;
                        if (nets.hasNext()) {
                            net = (ShadowNet) nets.next();
                            if (nets.hasNext()) {
                                System.err.println("DefaultShadowNetLoader: "
                                                   + url
                                                   + " contains more than one shadow net. Ignored.");
                            } else if (net.getName().equals(name)) {
                                // Transfer information about formalism
                                if (net.getCompilerFactory() == null) {
                                    net.setCompilerFactory(netSystem
                                                           .getDefaultCompilerFactory());
                                }


                                // Immediately exit the search
                                // loop and return the loaded net.
                                return net;
                            }
                        }
                        System.err.println("DefaultShadowNetLoader: " + url
                                           + " was empty!? Ignored.");
                        error = true;
                    } catch (Exception e) {
                        System.err.println("DefaultShadowNetLoader: " + url
                                           + " caused " + e);
                        if (PluginProperties.debug()) {
                            e.printStackTrace();
                        }
                        error = true;
                    }
                }
                ext++;
            }
        }
        return null;
    }

    public static class PathEntry {

        /**
         * A path to use when looking for nets. The path is stored using
         * OS-specific file system conventions.
         **/
        public final String path;

        /**
         * Indicates that the <code>path</code> component is considered to
         * be relative against the Java classpath. This means that this
         * <code>path</code> should not be looked up in the filesystem
         * directly, but used for a {@link ClassLoader#getResource} query.
         * The <code>CLASSPATH</code> keyword used in the String notation
         * of the net path has been removed from the path component.
         **/
        public final boolean isClasspathRelative;

        public PathEntry(String path, boolean isClasspathRelative) {
            this.path = path;
            this.isClasspathRelative = isClasspathRelative;
        }
    }
}